/* ==================================================
   network.c

   Ben Brillat
   Network Programming
   Project 3, TFTPd
   Due: 25 February 2000

   
   All network functions.


   Some code borrowed from Prof. Hollinger's UDP examples
   for Network Programming, Spring 2000, RPI.

   ================================================== */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <netdb.h>
#include <strings.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/tftp.h> //tftp defines


#include "network.h"
#include "parse.h"

/* ==================================================

   server - the main server function

   Parameters:
   initial_sock - the socket to initially read from

   Return value:
   negative on error 
   1 on success

   ================================================== */
int server(int initial_sock){

  int retval;
  int fd;
  struct sockaddr_in remote_sa;
  char buff[SEGSIZE+100];
  int length;
  short opcode;
  int mode;

  retval = read_data(initial_sock, 25, &remote_sa, buff, SEGSIZE);

  if(retval < 0){
    //error doing read_data for the request line
    return -40;
  }
  length = retval;
  
  //parse the request
  fd = parse_request(buff, length, &opcode, &mode);
  
  if(fd < 0){
    //error parsing request - send an error packet and die
    if(fd == -50)
      send_error(EBADOP, "Improperly formatted request", &remote_sa, -1);
    else if(fd == -51)
      send_error(EBADOP, "wrong opcode - not a request", &remote_sa, -1);
    else if(fd == -52)
      send_error(EBADOP, "No such mode", &remote_sa, -1);
    else if(fd == -53)
      send_error(ENOTFOUND, "Error opening file", &remote_sa, -1);
    else
      send_error(EBADOP, "Error parsing request", &remote_sa, -1);
    return -41;
  }
  
  //find out what we have to do
  if(opcode == RRQ){ //if we're reading...
    //find out if we have to do any conversion (Netascii mode [2])
    if(mode == 2){
      retval = htonetascii(fd);
      if(retval < 0){
	//something went wrong converting to netascii mode
	send_error(EACCESS, "Error converting to NETASCII", &remote_sa, -1);
 	return -42;
      }
      fd = retval; //save the new file descriptor
    }
    //now we can treat everything the same - binary.
    retval = send_file(fd, &remote_sa);
    if(retval < 0){
      //something went wrong sending the file....
    }
  }
  else if(opcode == WRQ){ //we're writing

    retval = receive_file(fd, &remote_sa, mode);
    
  }
 
  //we're done.  close the file
  retval = close(fd);
  if(retval < 0){
    perror("Error closing file");
  }
  
  if(opcode == RRQ && mode == 2){ //if we created a temp file
    snprintf(buff, 399, "%s/tftp-%d.netascii", P_tmpdir, getpid());
    retval = unlink(buff);
    if(retval < 0){
      perror("Error deleting temporary file");
    }
  }

  return 0;
} //end server()




/* ==================================================

  do_bind()

  bind to a port - must be given a sockaddr_in pointer (with memory already allocated!) and a port number (with memory!)

  use 0 as port number to have the number assigned.  

  return value:
  a file descriptor to your socket on success, negative number on error.
  

 ================================================== */

int do_bind(struct sockaddr_in *addr, int *port){

  int sock;
  int retval;
  int length;

  sock = socket(PF_INET, SOCK_DGRAM, 0);
  if(sock == -1){
    return -10; //error creating socket
  }

  addr->sin_family = AF_INET;
  addr->sin_port = htons(*port);
  addr->sin_addr.s_addr = htonl(INADDR_ANY);

  retval = bind(sock, (struct sockaddr *)addr, sizeof(*addr));

  if(retval != 0){
    return -11; //error doing bind
  }

  length = sizeof(*addr);
  if(getsockname(sock, (struct sockaddr *) addr, &length) < 0) { //fill in addr
    return -12; //error doing getsockname
  }
  
  //return port number.
  *port = ntohs(addr->sin_port);

  // fprintf(stderr, "do_bind() succeeded to UDP port: %d\n", *port); //DEBUG

  return sock;

} //end do_bind()


/* ==================================================

   read_data()

   Read data from a UDP socket - using recvfrom() & select()
   
   Parameters:
   sock - the socket to read from
   waittime - the time (in seconds) to wait for data on the socket
              set to -1 to wait forever, zero return immediately.
   sa - a sockaddr struct (which must have memory) that will be filled
        with info concerning the client that connected to our server.
   buff - where the data will be put. (must have memory!)
   maxlen - the maximum number of bytes that will fit in buff


   Return Value:
   negative number on error, the number of bytes read on success.

   ================================================== */

int read_data(int sock, int waittime, struct sockaddr_in *sa, 
	     char *buff, int maxlen){

  int retval;
  int length;

  retval = wait_udp(sock, waittime); //wait for data

  if(retval < 0){
    //an error happened
    return -20;
  }
  else if(retval == 0){
    //the timeout expired
    return -21;
  }
  //otherwise, there was data on our socket.  Read it.
  
  length = sizeof(*sa);
  

  retval = recvfrom(sock, buff, maxlen, 0, (struct sockaddr *)sa, &length);
  
  if(retval == -1){
    //an error happened during recvfrom
    return -22;
  }

  //if everything worked, return the number of bytes read.
  return retval;
  
} //end read_data()

/* ==================================================

   wait_udp - wait for a connection using select()

   Parameters:
   sock - the socket to wait on
   waittime - the time to wait: -1 to wait forever

   Return value:
   negative on error, zero if timeout expired, 
   1 if the descriptor has data waiting.

   ================================================== */

int wait_udp(int sock, int waittime){
  fd_set fds;
  struct timeval tv;
  int retval;

  FD_ZERO(&fds);
  FD_SET(sock, &fds);

  tv.tv_sec = waittime;
  tv.tv_usec = 0;

  //now wait for the data
  if(waittime == -1){
    //wait forever
    retval = select(sock+1, &fds, NULL, NULL, NULL);
  }
  else{
    //wait the specified time, or not at all.
    retval = select(sock+1, &fds, NULL, NULL, &tv);
  }

  if(retval < 0){
    //an error happened
    return -30;
  }
  
  return retval;

} //end wait_udp()




/* =================================================

   send_file - send a file to the given remote computer using TFTP

   Parameters:
   fd - the file descriptor for the file to send
   cli_sa - the sockaddr_in for the remote computer.

   Return Values:
   bytes written on success,
   negative number on error

   ============================================== */
int send_file(int fd, struct sockaddr_in *cli_sa){

  int serv_port; //also known as the TID of the server
  int cli_port; //also known as the TID of the client
  struct sockaddr_in serv_sa;
  int retval;
  int sock;
  int len;
  int retval2;
  char buff[SEGSIZE+100];
  char packet[SEGSIZE+100];
  short block;
  short opcode;
  struct sockaddr_in temp_sa;
  int i;
  int numbytes;

  cli_port = ntohs(cli_sa->sin_port); //get the remote port
  
  //bind to the new port
  sock = do_bind(&serv_sa, &serv_port);

  //send the file
  block = 1;
  opcode = DATA;
  numbytes = 0;
  retval = read(fd, buff, SEGSIZE); //read 512 bytes of file...
  while(retval >= 0){

    len = assemble_packet(&opcode, &block, buff, retval, packet); //build the packet
    i = 0;
    do{ // as long as this packet hasn't been ack'd

      retval2 = udp_send(sock, cli_sa, packet, len); //send the packet
      if(retval2 != len){
	//something went wrong sending
     	send_error(EACCESS, "Error sending packet", cli_sa, sock);
	return -71;
      }
      
      //now wait for an ACK
      retval2 = read_data(sock, 25, &temp_sa, buff, SEGSIZE+99);
      if(retval2 < 0){
	//perror("Error reading ACK in send_file");
	return -72;
      }
    
      i++;
    }while((verify_ack(block, buff, retval2, &temp_sa, cli_sa) != 1) 
	   && (i < 6));

    if(i >= 5){
      //too many tries re-sending data
      return -73;
    }

    numbytes += retval;
    if(retval < SEGSIZE){
      //we're all done
      break;
    }
    retval = read(fd, buff, SEGSIZE); //read the next segment
    block++;
  }
  if(retval < 0){
    //error doing read()
    send_error(EACCESS, "Error reading file", cli_sa, sock);
    return -70;
  }

  return numbytes;

} //end send_file()


/* ==================================================

   assemble_packet - build a tftp packet

   Parameters:
   short *opcode - a pointer to the TFTP opcode
   short *block - a pointer to the block number
   char *buff - a pointer to the stuff to put at the end of the packet
   int len - the number of bytes worth of buff to put at the end of the packet.
   char *packet - a pointer that will contain the whole packet (must have mem!)

   Return value:
   the total length of the packet on success
   a negative number on error

   ================================================== */
int assemble_packet(short *opcode, short *block, char *buff, int len,
		    char *packet){

  short n_opcode;
  short n_block;
  char *ptr;

  n_opcode = htons(*opcode); //put in network byte order
  n_block = htons(*block); //put in network byte order
  bzero(packet, SEGSIZE+99);
  
  ptr = packet;
  memcpy(ptr, &n_opcode, 2); //copy the opcode
  ptr += 2;
  memcpy(ptr, &n_block, 2); //copy the block number
  ptr += 2;
  memcpy(ptr, buff, len); //copy the other stuff

  return (len+4); //return the total length of this thing.

} //end assemble_packet



/* ==================================================

   udp_send() - send data out a udp socket

   Parameters:
   sock - the socket
   dest_sa - a pointer to the sockaddr_in describing the destination
   buff - a pointer to the stuff to send
   len - the length of stuff to send

   Return value:
   the number of bytes sent on success
   a negative number on error

   ================================================== */
int udp_send(int sock, struct sockaddr_in *dest_sa,
	     char *buff, int len){

  int numbytes=0;
  char *ptr;
  int sent;

  ptr = buff;
  while(numbytes < len){
    sent = sendto(sock, ptr, len-numbytes, 0,
		  (struct sockaddr *) dest_sa, sizeof(*dest_sa));
    if(sent < 0){
      //error sending data
      perror("Error in udp_send()");
      return -80;
    }
    numbytes += sent;
    ptr = buff+numbytes;
  }
  
  return numbytes;

}//end udp_send()


/* ==================================================

   verify_ack() - verify that the ack was in fact for the correct block.

   Parameters:
   block - the block that we should have
   buff - a pointer to the data that was read
   len - the length of the data that was read
   read_sa - a pointer to the sockaddr_in that was read in
   expected_sa - the sockaddr_in* that the ack should have come from


   Return value:
   1 if the ack matches
   negative number if it doesn't

   ================================================== */
int verify_ack(short block, char *buff, int len, struct sockaddr_in *read_sa,
	       struct sockaddr_in *expected_sa){

  short read_block;
  short read_opcode;
  short read_block_net;
  short read_opcode_net;
  char *ptr;

  ptr = buff;

  memcpy(&read_opcode_net, ptr, 2); //get the opcode
  ptr = buff+2;
  memcpy(&read_block_net, ptr, 2); //get the block

  read_opcode = ntohs(read_opcode_net);
  read_block = ntohs(read_block_net);

  if(read_opcode != ACK){
    //wrong opcode
    send_error(EBADOP, "Not an ack opcode", read_sa, -1);
    return -80;
  }
  if(read_block != block){
    //wrong block
    return -81;
  }
  if(len != 4){
    //why were there more than 4 bytes in this packet???
    send_error(EUNDEF, "More than 4 bytes in ack packet", read_sa, -1);
    return -82;
  }
  if(ntohs(read_sa->sin_port) != ntohs(expected_sa->sin_port)){
    //the TID was wrong...

    //SEND ERROR PACKET HERE - to the client that sent us this wrong packet
    send_error(EBADID, "Packet from wrong source", read_sa, -1);
    return -83;
  }
  return 1; //if all that is true, return true.
} //end verify_ack


/* ==================================================

   send_error() - send an error packet

   Parameters:
   short tftperr - the error
   char *mesg - a null-terminated string with the message
   struct sockaddr_in *dest_sa - the destination of the error packet
   int sock - the socket to send it out (-1 to have the computer pick)


   Return value:
   1 if sent successfully
   negative number if not

   ================================================== */
int send_error(short tftperr, char *mesg, struct sockaddr_in *dest_sa,
	       int sock){

  int len;
  int retval;
  short opcode = ERROR;
  char buff[SEGSIZE+100];
  int tempsock;

  len = assemble_packet(&opcode, &tftperr, mesg, strlen(mesg)+1, buff);
  if(len < 0){
    //error assembling packet
    return -90;
  }
  
  if(sock != -1){
    retval = udp_send(sock, dest_sa, buff, len);
  }
  else{
    tempsock = socket(PF_INET, SOCK_DGRAM, 0); //generate a socket if we
    //need to use a different port
    retval = udp_send(tempsock, dest_sa, buff, len);
  }
  if(retval != (strlen(mesg)+5)){
    //wrong number of bytes written
    return -91;
  }
  
  return 1;
  
}//end send_error()


/* ==================================================

   send_ack() - send an ack packet

   Parameters:
   short block - the block number to ack
   struct sockaddr_in *dest_sa - the destination of the error packet
   int sock - the socket to send it out (-1 to have the computer pick)

   Return value:
   1 if sent successfully
   negative number if not

   ================================================== */
int send_ack(short block, struct sockaddr_in *dest_sa, int sock){

  int len;
  int retval;
  short opcode = ACK;
  char buff[SEGSIZE+100];
  int tempsock;
  char fake[2];

  len = assemble_packet(&opcode, &block, fake, 0, buff);
  if(len < 0){
    //error assembling packet
    return -100;
  }
  
  if(sock != -1){
    retval = udp_send(sock, dest_sa, buff, len);
  }
  else{
    tempsock = socket(PF_INET, SOCK_DGRAM, 0); //generate a socket if we
    //need to use a different port
    retval = udp_send(tempsock, dest_sa, buff, len);
  }
  if(retval != 4){
    //wrong number of bytes written
    return -101;
  }
  
  return 1;
  
}//end send_ack()


/* =================================================

   receive_file - receive a file from the given remote computer using TFTP

   Parameters:
   fd - the file descriptor for the file to send
   cli_sa - the sockaddr_in for the remote computer.
   mode - the mode of this transfer (2-netascii or 1-octet)

   Return Values:
   bytes read to the file total returned on success,
   negative number on error

   ============================================== */
int receive_file(int fd, struct sockaddr_in *cli_sa, int mode){

  short block;
  int cli_port;
  int sock;
  struct sockaddr_in serv_sa;
  int serv_port;
  char buff[SEGSIZE+100];
  int numbytes;
  struct sockaddr_in temp_sa;
  int retval;
  int i = 0;
  int totalbytes = 0;

  cli_port = ntohs(cli_sa->sin_port); //get the remote port
  sock = do_bind(&serv_sa, &serv_port); //bind to a port here

  //ack the 0 packet (the request)
  block = 0;
  send_ack(0, cli_sa, sock);

  //receive the file
  numbytes = read_data(sock, 25, &temp_sa, buff, SEGSIZE+100);
  while(numbytes > 0){ //so long as we're reading in 516 bytes
    block++;
    retval = parse_data(buff, numbytes, mode, fd, block, &temp_sa, cli_sa);
    while(retval == -120 && i < 5){
      //read again, as long as we have the wrong thing.
      numbytes = read_data(sock, 25, &temp_sa, buff, SEGSIZE+100);
      retval = parse_data(buff, numbytes, mode, fd, block, &temp_sa, cli_sa);
      i++;
    }
    if(retval != (numbytes-4)){
      //error parsing data & writing to file
      //send_error(EUNDEF, "Error writing file", &temp_sa, -1);
      return -110;
    }
    
    //so, we read their block fine.  send an ack.
    send_ack(block, cli_sa, sock);
    totalbytes += retval;

    if(numbytes < (SEGSIZE+4)){
      //we read the last packet.
      break;
    }
    numbytes = read_data(sock, 25, &temp_sa, buff, SEGSIZE+100);
  }
  if(numbytes < 0){
    //there was an error.
    send_error(EUNDEF, "Error writing file", &temp_sa, -1);
    return -111;
  }

  return totalbytes;

}//end receive_file()

/* =================================================

Random Notes:

  struct in_addr {
   unsigned long int s_addr;
   }

   struct sockaddr_in {
     u_char  sin_len;
     u_short sin_family;
     u_short sin_port;
     struct  in_addr sin_addr;
     char    sin_zero[8];
   }

  fprintf(stderr, "%d bytes read from %s:%d\n%s\n", retval, 
 	  inet_ntoa(remote_sa.sin_addr), ntohs(remote_sa.sin_port), buff);


   ============================================== */

