/*
  parse.c

  Ben Brillat
  
  parse routines for TFTPd
 */

#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/tftp.h>

#include "parse.h"
#include "network.h"

/* ==================================================

   parse_request - parse the initial request (RRQ or WRQ)

   Parameters:
   char *buff - a pointer to the string that was sent
   int length - the length of this string
   short *opcode - the opcode (must have memory)
   int *mode - the mode of transfer - 1 for binary, 2 for netascii - must have memory

   Return value:
   negative on error 
   positive number that is a file descriptor to the requested file on success


            2 bytes     string    1 byte     string   1 byte
            ------------------------------------------------
           | Opcode |  Filename  |   0  |    Mode    |   0  |
            ------------------------------------------------
   ================================================== */
int parse_request(char *buff, int length, short *opcode, int *mode){

  char filename[SEGSIZE];
  char modestr[10];
  short *s;
  char *c;
  int retval;
 
  //first, verify that this thing ends with a \0
  if(buff[length-1] != 0){
    //we don't have a properly formatted string
    return -50;
  }

  //second, get the opcode
  s = (short *)buff;
  *opcode = ntohs(*s);

  if(*opcode != RRQ && *opcode != WRQ){
    //this is neither a read nor a write request..
    return -51;
  }
  
  //then, get the filename
  c = buff+2;
  strncpy(filename, c, SEGSIZE);

  //last, get the mode
  c = c+strlen(filename)+1;
  strncpy(modestr, c, 10);
  if(strncasecmp(modestr, "netascii", 8) == 0){
    *mode = 2;
  }
  else if(strncasecmp(modestr, "octet", 5) == 0){
    *mode = 1;
  }
  else{
    //no such mode of transfer...
    return -52;
  }
  
  //DEBUG
  // fprintf(stderr, "op: %hd, file: %s, mode: %s.\n", *opcode, filename, modestr);

  //now open the file.
  if(*opcode == RRQ){
    //we have a read request
    retval = open(filename, O_RDONLY);
  }
  else{
    //we have a write request
    retval = open(filename, O_WRONLY | O_CREAT | O_TRUNC,
		  S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
  }

  if(retval == -1){
    //an error occured doing the open()
    return -53;
  }


  return retval; //return the file descriptor

} //end parse_request()


/* ==================================================

   htonetascii - convert from UNIX text file format to netascii

   Parameters:
   orig - file descriptor to original file

   Return Value:
   file descriptor to new file in netascii format on success
   negative value on error.

 ================================================== */
int htonetascii(int orig){

  int na_fd;
  char filename[400];
  int retval;
  char c;
  char d=0;

  snprintf(filename, 399, "%s/tftp-%d.netascii", P_tmpdir, getpid());
 
 
  na_fd = open(filename, O_RDWR | O_CREAT | O_TRUNC, 
	       S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);

  //now, read the UNIX-format file and output to a NETASCII-format file.
  retval = read(orig, &c, 1);
  while(retval > 0){
    if(c == '\n' && d != '\r'){ //make sure we have LF, not CRLF
      retval = write(na_fd, "\r\n", 2); //if we have a LF, write CRLF
    }
    else{
      retval = write(na_fd, &c, 1); //otherwise, write the character as read.
    }
    if(retval < 1){
      //something went wrong doing the write
      return -60;
    }
    //move on to the next letter.
    d = c;
    retval = read(orig, &c, 1);
  }

  if(retval < 0){
    //an error happened reading the original file
    close(na_fd);
    close(orig);
    perror("Error reading file in htonetascii");
    return -61;
  }
  //otherwise, just close the original file.
  retval = close(orig);
  if(retval < 0){
    perror("Error closing original file");
    return -62;
  }

  //move the offset to the beginning of the netascii file
  retval = lseek(na_fd, 0, SEEK_SET);
  if(retval != 0){
    //we're not at the beginning of the file for some reason..
    perror("unable to lseek to the beginning of the temp file");
    return -63;
  }
  
  //everything worked!
  return na_fd;

} //end htonetascii


/* ==================================================

   parse_data - parse the data coming in for write
                and write it to the file in fd

   Parameters:
   char *buff - a pointer to the string that was sent
   int length - the length of this string
   int mode - the mode of transfer - 1 for binary, 2 for netascii
   short block - the block that we are expecting this to be.
   struct sockaddr_in *from_sa - the sockaddr_in that this packet came from
   struct sockaddr_in *expected_sa - where we expected the packet to come from

   Return value:
   negative on error 
   number of bytes written to the file on success

                   2 bytes     2 bytes      n bytes
                   ----------------------------------
                  | Opcode |   Block #  |   Data     |
                   ----------------------------------
                        Figure 5-2: DATA packet

   ================================================== */
int parse_data(char *buff, int length, int mode, int fd, short block,
	       struct sockaddr_in *from_sa,
	       struct sockaddr_in *expected_sa){

  short rcvd_block;
  short rcvd_block_net;
  short opcode;
  short opcode_net;
  char *ptr;
  int retval;
  int i;
  int totalbytes=0;

  ptr = buff;
  memcpy(&opcode_net, ptr, 2);
  ptr = buff+2;
  memcpy(&rcvd_block_net, ptr, 2);
  ptr = buff+4;
  opcode = ntohs(opcode_net);
  rcvd_block = ntohs(rcvd_block_net);

  if(ntohs(from_sa->sin_port) != ntohs(expected_sa->sin_port)){
    //the TID was wrong...
    send_error(EBADID, "Packet from wrong source", from_sa, -1);
    return -122;
  }
  if(opcode != DATA){
    send_error(EBADOP, "Expecting a data packet", from_sa, -1);
    return -121;
  }
  if(rcvd_block != block){
    //wrong block
    return -120;
  }
  retval = 0;
  if(mode == 2){ //then NETASCII - remove CR's and leave only LF's
    for(i=0; i<(length-4); i++){
      if(ptr[0] != '\r'){
	retval = write(fd, ptr, 1);
	if(retval < 0){
	  break;
	}
      }
      ptr++; //move to the next character
      totalbytes++;
    }
  }
  else{
    retval = write(fd, ptr, length-4);
    totalbytes+=retval;
  }

  if(retval < 0){
    //something went wrong writing the file
    perror("Error writing file in parse_data()");
    send_error(ENOSPACE, "Unable to write", from_sa, -1);
    return -123;
  }
  
  return totalbytes;

}//end parse_data
